@@ -1,5 +1,5 @@
 import numpy as np
-from ._base import DMPBase
+from ._base import DMPBase, WeightParametersMixin
 from ._forcing_term import ForcingTerm
 from ._canonical_system import canonical_system_alpha
 
@@ -296,7 +296,7 @@ except ImportError:
     pass
 
 
-class DMP(DMPBase):
+class DMP(WeightParametersMixin, DMPBase):
     """Dynamical movement primitive (DMP).
 
     Implementation according to
@@ -467,27 +467,6 @@ class DMP(DMPBase):
             alpha_z=self.forcing_term.alpha_z, allow_final_velocity=allow_final_velocity)
         self.configure(start_y=start_y, goal_y=goal_y)
 
-    def get_weights(self):
-        """Get weight vector of DMP.
-
-        Returns
-        -------
-        weights : array, shape (n_dims * n_weights_per_dim,)
-            Current weights of the DMP.
-        """
-        return self.forcing_term.weights.ravel()
-
-    def set_weights(self, weights):
-        """Set weight vector of DMP.
-
-        Parameters
-        ----------
-        weights : array, shape (n_dims * n_weights_per_dim,)
-            New weights of the DMP.
-        """
-        self.forcing_term.weights[:, :] = weights.reshape(
-            -1, self.n_weights_per_dim)
-
 
 def dmp_transformation_system(Y, V, alpha_y, beta_y, goal_y, goal_yd, goal_ydd,
                               execution_time):
